#include "MouseIO.h"
#include "MouseCell.h"
#include "Chromosome.h"
#include "CellData.h"

#include <stdio.h>

#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <qstring.h>

using namespace std;

// ================================================================
// GenomeMixer    (c) 2003, Alex Williams
// All rights reserved.
// Send email correspondence to Robert W. Williams <rwilliam@nb.utmem.edu>
// at the University of Tennesse Health Science Center

// This project was funded jointly by the National Institute of Mental Health, National Institute on Drug Abuse,
// and the National Science Foundation (award P20-MH 62009 to KFM and RW)

// This source code may be used according to the terms of the GNU General Public License (GPL).
// See www.gnu.org for specifics. Essentially, the source may be freely used as long as the source code
// for whatever program the source is also made freely available. If you want to use this source
// in a closed-source program, you should email us for licensing options.

// Disclaimer: This code comes with no warranty at all, and there is no guarantee that it will
// work as claimed in the documentation. Hopefully, it will be useful anyway.

// Note that GenomeMixer makes use of the Trolltech Qt toolkit for its window-managing needs.
// There is very little use of Qt data structures (aside from QString) in the actual processing code,
// but **you will need to have Qt installed in order to compile this code.** Check <www.trolltech.com>
// for details; there should be a free GPL-software-only version of Qt available on the Trolltech
// web site.

// Any redistributions of this software in code or binary form must include the
// above copyright notice, licensing information, disclaimer, and this notice itself.
// Basically, all the comments at the top of this file must be retained.
// ================================================================



// ========================================================================================================
// ========================================================================================================
void MouseIO::readBreedingSchemeFile(string fileToRead, GenomeMixer *mixer) {	
	try { 
		string r;
		int numChr = 0, numPurebreds = 0, numGenerations = 0, numCages = 0;
		bool hasFemaleSexChr = false, hasMaleSexChr = false;
		
		if (mixer == NULL) {
			throw new string("Error: readBreedingScheme file was called with a non-existent GenomeMixer.");
		}
		
		// *******************************************
		// First, go through and count to see how many CHR there are, how
		// many purebreds, etc...
		ifstream file;
		file.open(fileToRead.c_str(), ios::in);
		if (file.fail() || !file.is_open()) {
			throw new string(string("MouseIO: Error occurred while trying to open breeding scheme file \"") + fileToRead + string("\" for reading."));
		}
		
		while (true) { // loop until we get a read failure
			file >> r;
			if (file.fail() || r == "<END_DATA>") { // <-- yes, the == is legal for string comparison
				// stop reading when we get a failure (no more to read), or we read <END_DATA>
				break;
			}
			// kSTRCMP_EQUAL is just the number 0
			if (r == "<CHR>") {
				numChr++;
			} else if (r == "<FEMALE_SEX_CHR>") {
				numChr++;
				hasFemaleSexChr = true;
			} else if (r == "<MALE_SEX_CHR>") {
				numChr++;
				hasMaleSexChr = true;
			} else if (r == "<PUREBRED>") {
				numPurebreds++;
			} else if (r == "<GENERATIONS") {
				file >> numGenerations;
			} else if (r == "<CAGE") {
				numCages++;
			}
			//cout << "[" << r << "]"; // << endl;
		}
		file.close();
		
		//cout << endl;
		//cout << "numChr is " << numChr << ", numPurebreds is " << numPurebreds <<  ", numGenerations is " << numGenerations << ", numCages is " << numCages << "." << endl;
		//cout << "hasFemaleSexChr is "  << hasFemaleSexChr << ", hasMaleSexChr is " << hasMaleSexChr << endl;
		
		// *******************************************
		// Now that we have numbers for everything, go through and actually
		// read in the data (second pass through file).
		// But first, we need to allocate space.
		Chromosome::setNumberOfChromosomes(numChr); // <-- sets the number of chromosomes
		mixer->setTableSize(numCages, numGenerations);
		mixer->clearAllParentalStrains();
		
		// Done allocating memory for the data to be read
		
		file.clear(); // clear any "failure" status
		file.open(fileToRead.c_str(), ios::in);
		if (file.fail()) {
			throw new string(string("MouseIO: Error opening breeding scheme file \"") + fileToRead + string("\" for reading the second pass\n(strange, since we opened it just fine a second ago)"));
		}
		
		int numCagesReadSoFar = 0;
		int chrsReadSoFar     = 0;
		while (true) { // loop until we get a read failure
			file >> r;
			//file.getline(file, r, '\t');
			if (file.fail() || (r == "<END_DATA>")) {
				// stop reading when we get a failure (not really a "failure," just no more to read),
				// or when we read <END_DATA>, which indicates that the rest of the file is comments
				break;
			}
			
			if (r == "<NAME>") {
				file >> mixer->name;
				
			} else if (r == "<VERSION>") {
				file >> mixer->version; // floating-point number
				
			} else if (r == "<CHR>") {
				float float_MB, float_cM;
				string chrName;
				file >> chrName;
				file >> float_MB;
				file.ignore(3); // MB
				file >> float_cM;
				Chromosome::addChrData(chrName, float_MB, float_cM);
				chrsReadSoFar++;
				
			} else if (r == "<FEMALE_SEX_CHR>") {
				float   float_MB, float_cM;
				string femaleChrName;
				file >> femaleChrName;
				file >> float_MB;
				file.ignore(3); // MB
				file >> float_cM;
				Chromosome::addChrData(femaleChrName, float_MB, float_cM);
				chrsReadSoFar++;
				
			} else if (r == "<MALE_SEX_CHR>") {
				float   float_MB, float_cM;
				string maleChrName;
				file >> maleChrName;
				file >> float_MB;
				file.ignore(3); // MB
				file >> float_cM;
				Chromosome::addChrData(maleChrName, float_MB, float_cM);
				chrsReadSoFar++;
				
			} else if (r == "<DEFAULT_LITTER_SIZE>") {
				file >> mixer->defaultLitterSize;
				
			} else if (r == "<DEFAULT_LITTER_STDEV>") {
				file >> mixer->defaultLitterStDev;
			} else if (r == "<ABSOLUTE_INTERFERENCE>") {
			  file >> mixer->noRecombDistBases;
			  
			} else if (r == "<RECOMBINATION_MODEL>") {
			  file >> mixer->recombModel;
			  
			} else if (r == "<PUREBRED>") {
				string purebredName, purebredSymbol; // the purebred name
				file >> purebredName >> purebredSymbol;
				mixer->addParentalStrain(purebredName, purebredSymbol);
			} else if (r == "<GENERATIONS") {
				// We don't have anything else to do for this line (we already read in the number of generations earlier, above)
				
			} else if (r == "<CAGE") {
				// Read the strings into the array.
				if (numCagesReadSoFar >= mixer->numCages()) {
					cerr << "Num cages read so far: " << numCagesReadSoFar
						<< ", mixer->numCages(): " << mixer->numCages() << endl;
					if (numCagesReadSoFar >= mixer->numCages()) {
						throw new string("Error in reading the cages: we must have miscounted\nthe number of cages, or the file changed while\nwe were reading it.");
					}
				}
				
				int cageNum;
				string s;
				file >> cageNum >> s; // reads in a number and an angle bracket (something like "16>" (the current cage number) )
				
				if (cageNum != numCagesReadSoFar+1) {
					cerr << "Warning: It looks like the cages are numbered non-sequentially in the input file." << endl;
				}
				
				bool inCell = false; // whether or not we are reading a cell (in the "tab" region between cells)
				int  gen = 0; // which generation (column) we are reading from the table
				char c;
				string readInString;
				do {
					c = file.get();
					if (c == '\"') {
						if (inCell) {
							mixer->setBreedingTableElem(numCagesReadSoFar, gen, readInString);
							mixer->setCellData(numCagesReadSoFar, gen, MouseIO::stringToCellData(readInString, mixer));
							readInString = "";
							gen++;
						}
						inCell = !inCell;
						// ever time we encounter a ", it changes our status of 
						// whether or not we are in a cell, or in the non-quoted region between cells
						
					} else if (inCell) {
						assert(gen < mixer->numGenerations());
						readInString += c;
					}
				} while (c != '\n' && c != '\r' && !file.eof());
				numCagesReadSoFar++;
			}
			//cout << "[" << r << "]"; // << endl;
		} // <-- end of "while (true)"
		
		file.close(); // close the config/table file after getting the data from it
		
		assert(numChr == Chromosome::numChr()); // make sure we read everything properly
		
	} catch (const string *e) {
		throw e;
	}
}


// ====================================================================================================================
// ====================================================================================================================

CellData* MouseIO::stringToCellData(string cell, GenomeMixer *mixer) {
	try {
		// This function can allocate memory!
		
		assert(mixer != NULL);
		bool commentedOut = false, exportThisLitter = false;
		
		if (cell == "" || cell == NULL) {
			return NULL; // there is simply no data in this cell (this is OK)
			
		} else if (cell.size() <= 3) {
			throw new string(string("Sorry, the cell \"") + cell + string("\" is not a valid cell."));
			// not enough data in this cell to decode it
		}  
		
		if (cell[0] == '#') {
			commentedOut = true;
			cell = cell.substr(1, cell.size() - 1); // clip off the #
		}
		
		if (cell.substr(0, 3) == string("exp")) {
			exportThisLitter = true;
			cell = cell.substr(4, cell.size() - 4); // 4, because we don't want "exp " <-- including the space
		}
		
		unsigned int temp = cell.find_first_of(' ', 0);
		if (temp == string::npos) {
			throw new string(string("Sorry, the cell \"") + cell + string("is not valid.\nWe couldn't read in the female parent."));
		}
		string femaleSourceStr = cell.substr(0, temp); // reads up until the first space, into firstSpecStr (so now it will be "B29" or something)
		cell = cell.substr(temp + 1, cell.size() - temp - 1); // now cut off that part of the string
		
		temp = cell.find_first_of('x', 0);
		if (temp == string::npos) {
			throw new string(string("Sorry, the cell \"") + cell + string("is not valid.\nWe couldn't read in the 'x' that\nseparates the female and male parents."));
		}  
		cell = cell.substr(temp + 2, cell.size() - temp - 2); // now cut off that part of the string, and the space afterwards
		
		temp = cell.find_first_of(' ', 0);
		if (temp == string::npos) {
			throw new string(string("Sorry, the cell \"") + cell + string("is not valid.\nWe couldn't read in the male parent."));
		}
		string maleSourceStr = cell.substr(0, temp); // reads up until the first space, into secondSpecStr (so now it will be "p(AA)" or something)
		cell = cell.substr(temp + 1, cell.size() - temp - 1); // now cut off that part of the string
		
		// Now we should have something like "r=25 m=11 f=12" remaining
		
		int num = 0, fem = 0, male = 0;
		unsigned int found_rand  = cell.find_first_of("rR", 0); // <-- it will be case-insensitive (finds first of either r or R)
		unsigned int found_fem  = cell.find_first_of("fF", 0);
		unsigned int found_male = cell.find_first_of("mM", 0);
		
		if (found_rand != string::npos) {
			sscanf( cell.substr(found_rand, cell.size() - found_rand).c_str(), "r=%d", &num);
		}
		if (found_fem != string::npos) {
			sscanf( cell.substr(found_fem, cell.size() - found_fem).c_str(), "f=%d", &fem);
		}
		if (found_male != string::npos) {
			sscanf( cell.substr(found_male, cell.size() - found_male).c_str(), "m=%d", &male);
		}
		
		int maleCage, maleGen, femCage, femGen, femPurebred, malePurebred;
		
		MouseIO::parentStrDecode(femaleSourceStr, mixer, &femCage , &femGen, &femPurebred);
		MouseIO::parentStrDecode(maleSourceStr  , mixer, &maleCage, &maleGen, &malePurebred);
		//cout << "num is " << num << ", fem is " << fem << ", male is " << male << endl;
		CellData *theCellData = new CellData(commentedOut, exportThisLitter, num, fem, male);		
		assert(theCellData != NULL);
		
		if (femPurebred == -1) {
			// so the mother isn't a purebred
			theCellData->setMother(femCage, femGen);
		} else {
			// ok, it's a purebred
			theCellData->setMotherPurebred(femPurebred);
		}
		
		if (malePurebred == -1) {
			// so the father isn't a purebred
			theCellData->setFather(maleCage, maleGen);
		} else {
			// ok, it's a purebred
			theCellData->setFatherPurebred(malePurebred);
		}
		
		// no standard deviation support yet!!    
		return theCellData; // a pointer, note
		
	} catch (const string* e) {
		throw e;
	}
}

// ====================================================================================================================
// ====================================================================================================================
void MouseIO::parentStrDecode(string stringItem, GenomeMixer *mixer, int *cagePtr, int *genPtr, int *purebredPtr) {
	// Sets purebredPtr to -1 if this stringItem does *not* indicate a 
	// purebred strain. In that case, it sets cagePtr and genPtr to the
	// values of the cage & generation of the mouse.
	
	// If this does turn out to be a purebred, then this function sets
	// purebredPtr to the index of whichever purebred strain it indicates.
	// (valid: 0 through whatever the maximum is)
	
	try {
		if (mixer == NULL) { 
			throw new string("Programming Error: MouseIO::parentStrDecode.\nCan't decode: the GenomeMixer pointer is NULL.");
		} else if (!cagePtr || !genPtr || !purebredPtr) {
			throw new string("Programming Error: MouseIO::parentStrDecode.\nCan't decode: at least one of cagePtr, genPtr, and purebredPtr is NULL.");
		} else if (stringItem == NULL) {
			throw new string("Programming Error: parentStrDecode was called with a NULL stringItem.");
		}
		
		// Used for filling in a CellData
		// This is a stripped-down version of "decodeOneSide," below. Check that for better comments.
		if (stringItem.substr(0,2) == "p(") { // p(XYZ) indicates "get the parental/purebred strain named XYZ"
			if (stringItem[stringItem.size() - 1] != ')') {
				throw new string("This breeding scheme cell has a malformed p(StrainName)\n(it was probably missing the close-parenthesis, or else\nthere were extra characters following the close parenthesis).\nTo indicate a purebred, you need a lower-case p, then an\nopen-parenthesis, then the strain name,\nand then a close-parenthesis.\n\nAlso, there cannot be any extra trailing characters,\nor you will get this error even if the part\ninside parentheses is okay.");
			}
			stringItem = stringItem.substr(2, stringItem.size() - 3); // remove "p(" and ")"
			for (int i = 0; i < mixer->numParentalStrains(); i++) {	
				if (mixer->getPurebredName(i) == stringItem) {
					*purebredPtr = i;
					return;  // got it: this is the purebred we were looking for
				}
			}
			// In case we didn't find what we were looking for, we get to here...
			throw new string(string("Error: In the breeding scheme file, there was a request for\nan invalid parental strain, \"") + stringItem + string("\".\nIt should be the name of a strain, in a format like (\"BXD37\").\nCheck the Purebreds tab to see a list of valid strains."));
			
		} else {
			char sourceCellChar1 = 0, sourceCellChar2 = 0; // example: A, example: B, or ' ' (if there is no second char)
			int sourceCellGen, sourceCellCage;
			unsigned int found_digit = stringItem.find_first_of("0123456789", 0); // <-- finds the index of the first digit in "stringItem"	
			if (found_digit == 1) { // Single letter to read ("B294"). Generation is specified by letter, cage by number.
				sscanf(stringItem.c_str(), "%c", &sourceCellChar1);
				sourceCellGen = MouseIO::generationFromCellLetters(sourceCellChar1, sourceCellChar2);
				stringItem = stringItem.substr(1, stringItem.size() - 1); // remove that one letter from the front	    
			} else if (found_digit == 2) { // Two letters to read ("AZ32021").
				sscanf(stringItem.c_str(), "%c%c", &sourceCellChar1, &sourceCellChar2);
				sourceCellGen = MouseIO::generationFromCellLetters(sourceCellChar1, sourceCellChar2);
				stringItem = stringItem.substr(2, stringItem.size() - 2); // remove those two letters from the front	    
			} else {
				throw new string(string("Error reading: the request for parent animal in cell \"") +
								 stringItem + 
								 string("\" is not valid.\nIt should be one or two letters, and then 1-3 digits, with no spaces (\"A9\")."));
			}
			sscanf(stringItem.c_str(), "%d", &sourceCellCage);
			sourceCellCage -= 1; // instead of row "1," we have row 0
			
			*genPtr = sourceCellGen;
			*cagePtr = sourceCellCage;
			*purebredPtr = -1;
		} // end decoding "HJ25" or whatever it is (a reference to a previous cell)
		
	} catch (const string *e) {
		throw e;
	}
}






// ====================================================================================================================
// ====================================================================================================================
bool MouseIO::decodeCell(MouseCellIngredients *ingredients, string cell, GenomeMixer *mixer) {
	try {
		assert(mixer != NULL);
		bool exportThisLitter = false;
		string cellCopy = cell;
		
		if (cellCopy.size() <= 3 || (cellCopy[0] == '#')) {
			// either: not enough data in this cell to decode, or it's commented out (with a #)
			return false;
		}
		
		//# Format:		Female x Male
		//# Below: decode the information from this cell of data
		//# p(AA) x p(BB)
		//# p(AA) x C2
		
		//cout << "MouseIO: Decoding: Got cell [" << cell << "]";thro
		
		// .substr(X, Y) works as follows: X is the position to start from, and Y is the number of chars to extract
		if (cellCopy.substr(0, 3) == string("exp")) {
			// See if this cell starts with "exp," which means it should be "exported"--written out to a file when we do our outputs.
			//cout << " (export)";
			exportThisLitter = true;
			cellCopy = cellCopy.substr(4, cellCopy.size() - 4); // 4, because we don't want "exp " <-- including the space
			//cout << "Now the cellCopy is [" << cellCopy << "]" << endl;
		}
		
		unsigned int temp = cellCopy.find_first_of(' ', 0);
		string femaleSourceStr = cellCopy.substr(0, temp); // reads up until the first space, into firstSpecStr (so now it will be "B29" or something)
		cellCopy = cellCopy.substr(temp + 1, cellCopy.size() - temp - 1); // now cut off that part of the string
		
		temp = cellCopy.find_first_of('x', 0);
		cellCopy = cellCopy.substr(temp + 2, cellCopy.size() - temp - 2); // now cut off that part of the string, and the space afterwards
		
		temp = cellCopy.find_first_of(' ', 0);
		string maleSourceStr = cellCopy.substr(0, temp); // reads up until the first space, into secondSpecStr (so now it will be "p(AA)" or something)
		cellCopy = cellCopy.substr(temp + 1, cellCopy.size() - temp - 1); // now cut off that part of the string
		
		// Now we should have something like "r=25 m=11 f=12" remaining
		
		int num = 0, fem = 0, male = 0;
		unsigned int found_rand  = cellCopy.find_first_of("rR", 0); // <-- case doesn't matter; finds first of 'r' or 'R'
		unsigned int found_fem  = cellCopy.find_first_of("fF", 0);
		unsigned int found_male = cellCopy.find_first_of("mM", 0);
		
		if (found_rand != string::npos) {
			// So the number of total mice to be generated for this cell has been explicitly set (r=)
			sscanf( cellCopy.substr(found_rand, cellCopy.size() - found_rand).c_str(), "r=%d", &num);
		}
		
		if (found_fem != string::npos) {
			// So the number of female mice to be generated for this cellCopy has been explicitly set (f=)
			sscanf( cellCopy.substr(found_fem, cellCopy.size() - found_fem).c_str(), "f=%d", &fem);
		}
		
		if (found_male != string::npos) {
			// So the number of male mice to be generated for this cellCopy has been explicitly set (m=)
			sscanf( cellCopy.substr(found_male, cellCopy.size() - found_male).c_str(), "m=%d", &male);
		}
		
		//cout << "For n: " << num << ", for m: " << male << ", for f: " << fem << endl;
		
		// Now we change the values of the passed-in MouseCellIngredients
		ingredients->mother = decodeOneSide(femaleSourceStr, kFEMALE, mixer);
		ingredients->father = decodeOneSide(maleSourceStr  , kMALE  , mixer);
		ingredients->litterSizeStdev  = 0;    // <-- no standard deviation support yet!!
		ingredients->numRandomSex = num;
		ingredients->numMales           = male;
		ingredients->numFemales       = fem;
		ingredients->shouldBeExported = exportThisLitter;  // <-- a boolean value
		return true; // indicating that we returned what we consider valid mouse data
		
	} catch (const string *e) {
		throw e;
	}
}



// ====================================================================================================================
// ====================================================================================================================
Mouse* MouseIO::decodeOneSide(string stringItem, bool whichSex, GenomeMixer *mixer) {
	try {
		string copiedItem(stringItem);
		
		// DECODE the format p(...)
		if (copiedItem.substr(0,2) == "p(") {
			// p(XYZ) indicates "get the parental/purebred strain named XYZ"
			if (copiedItem[copiedItem.size() - 1] != ')') {
				throw new string("Warning: While decoding the data in a cell in the breeding table, we found a malformed p(...) (it was missing the close-parenthesis)");
			}
			copiedItem = copiedItem.substr(2, copiedItem.size() - 3); // remove "p(" and ")"
			
			for (int i = 0; i < mixer->numParentalStrains(); i++) {
				if (mixer->getPurebredName(i) == copiedItem)
					return mixer->getPurebredPtr(i, whichSex); // <-- return our purebred mouse!
			}
			// In case we didn't find what we were looking for, we get to here...
			throw new string(string("Error: In the breeding scheme file, there was a request for the invalid parental strain \"") + copiedItem + string("\".\nMake sure it is defined, and remember that case matters,\nand that the strain can't have spaces, commas, or\nquotations marks in its name."));
		} else {
			// DECODE the form "HJ25"
			
			// We are looking for something NOT of the form p(...); so we want something like "B52" or "AZ999"	
			char sourceCellChar1 = 0; // example: A
			char sourceCellChar2 = 0; // example: B, or ' ' (if there is no second char)
			int sourceCellGen;    // example: 42
			int sourceCellCage;   // example: 2112
			unsigned int found_digit = copiedItem.find_first_of("0123456789", 0); // <-- finds the index of the first digit in "copiedItem"
			
			if (found_digit == 1) {
				// Single letter to read ("B294"). Generation is specified by letter, cage by number.
				sscanf(copiedItem.c_str(), "%c", &sourceCellChar1);
				sourceCellGen = MouseIO::generationFromCellLetters(sourceCellChar1, sourceCellChar2);
				copiedItem = copiedItem.substr(1, copiedItem.size() - 1); // remove that one letter from the front
				
			} else if (found_digit == 2) {
				// Two letters to read ("AZ32021"). Generation is specified by letter, cage by number.
				sscanf(copiedItem.c_str(), "%c%c", &sourceCellChar1, &sourceCellChar2);
				sourceCellGen = MouseIO::generationFromCellLetters(sourceCellChar1, sourceCellChar2);
				copiedItem = copiedItem.substr(2, copiedItem.size() - 2); // remove those two letters from the front
				
			} else {
				throw new string(string("Error reading: Something is wrong with the\nreference to another cell, in cell \"") + stringItem + string("\". Probably, this cell is trying to obtain data\nfrom a nonexistent cell (like \"AB93A\" or something)."));
			}
			
			// Generation is specified by letter, cage by number.
			sscanf(copiedItem.c_str(), "%d", &sourceCellCage);
			sourceCellCage -= 1; // instead of row "1," we have row 0 as the top
			
			// --------------
			if (Mouse::isFemale(whichSex)) {
				if (mixer->getMiceTable(sourceCellCage, sourceCellGen)->numFemaleMice() <= 0) {
					throw new string(string("Error: Attempted to get a female from cell \"") + stringItem + string("\", which has no females.\nYou may want to edit the breeding scheme file to ensure that at\nleast one female will be generated for this cell."));
				} else {
					// Ok, return a random female mouse now
					return mixer->getMiceTable(sourceCellCage, sourceCellGen)->selectRandomMouse(whichSex);
				}
				
			} else {
				if (mixer->getMiceTable(sourceCellCage, sourceCellGen)->numMaleMice() <= 0) {
					throw new string(string("Error: Attempted to get a male from cell \"" + stringItem + string("\", which has no males.\nYou may want to edit the breeding scheme file to ensure that at\nleast one male will be generated for this cell.")));
				} else {
					// Ok, return a random male mouse now
					return mixer->getMiceTable(sourceCellCage, sourceCellGen)->selectRandomMouse(whichSex);
				}
			}
			// --------------
			
		} // end decoding "HJ25" or whatever it is (a reference to a previous cell)
	} catch (const string *e) {
		throw e;
	}
}


// ======================================================================================================
// ======================================================================================================
string MouseIO::generationNumToLetters(int number) {
	// Generation number -> cell letters  
	// Given a number, like 0 or 27 or whatever, turns it into a corresponding spreadsheet-style lettering  
	// We only go up through 701. Hopefully no one will need more than 701 generations.
	//assert(number >= 0 && number <= 701); // 701 is "ZZ"
	
	QString returnStr;
	if (number > 25) {
		// AA = 26*(0 + 1) + 0 = 26     AB = 26*(0 + 1) + 1 = 27    ZZ = 26*(25 + 1) + 25 = 701
		returnStr  = char(number/26 + 'A' - 1); // turn 1 into 'A', etc (since this will be first called for 27='AA')
		returnStr += char(number%26 + 'A'); // turn 0 into 'A', etc
	} else {
		// A = 0               Z = 25
		returnStr = char(number%26 + 'A'); // turn 0 into 'A', etc
	}
	
	return returnStr;
}



// =====================================================================================================
// =====================================================================================================
int MouseIO::generationFromCellLetters(char letOne, char letTwo) {
	// A = 0      Z = 25
	// AA = 26*(0 + 1) + 0 = 26           AB = 26*(0 + 1) + 1 = 27         ZZ = 26*(25 + 1) + 25 = 701
	if (letTwo == 0) {
		// We only have one letter in the cell specification (like cell "A14")
		return	numFromChar(letOne);
	} else {
		// We have two letters in the specification (like cell "CD4" or "AZ423")
		return (26 * (numFromChar(letOne)+1) + numFromChar(letTwo));
	}
}

// =====================================================================================================
// =====================================================================================================
int MouseIO::numFromChar(char c) {
	// This turns a letter from A through Z into a corresponding integer, 0 to 25
	// A becomes 0, B becomes 1, ... Z becomes 25. Lower and upper case both work.
	// Only letters are valid input, so something like '*' would be rejected.
	if (c > 'Z') {
		c -= 32; // converts lower case ('a' == 97) into upper case ('A' == 65)
	}
	if (c < 'A' || c > 'Z') {
		cerr << "Error in converting the char \'" << c << "\'. We tried to convert something besides A-Z (a-z) into a number. Probably there is a bad entry in the breeding scheme cells." << endl;
		throw new string("Error in converting a char. Probably there is a bad entry in the breeding scheme cells.");
	}
	return (int(c) - 65); // gives 0 for A, 1 for B, ... , 25 for Z.
}


// ====================================================================================================================
// ====================================================================================================================
// ******** ******** ******** ******** ******** ******** ******** ******** ********
//				read marker location file
// ******** ******** ******** ******** ******** ******** ******** ******** ********
MarkerCollection* MouseIO::getMarkerDataFromFile(string markerFileLocation) {    
	
	try {
		
		MarkerCollection *newCollection = new MarkerCollection(10);
		ifstream file;
		
		file.open(markerFileLocation.c_str(), ios::in);
		if (file.fail() || !file.is_open()) {
			throw new string(string("MouseIO: Error occurred while trying to open the marker location file \"") + markerFileLocation + string("\" for reading."));
		}
		
		while (true) {
			string tempString;
			// This part just reads past the header data
			file >> tempString;
			if (file.fail()) {
				throw new string("Didn't find valid data in the marker file.");
				
			} else if (tempString == "<START") {
				break; // go on to the actual reading of marker data
			}
		}
		
		
		while (true) { // loop until we get a read failure (probably meaning "end of file")	
			// Ok, we got past the comments...
			string   s;
			string   markerNameStr, chrNumStr;
			long    locInBases;
			float   locIn_cM;
			int     interpretedChrIndex; // all the chromosome indices are one less than the number we read in from the file. Chr 1 is index 0, for example.
			
			file >> s;		// not the same as getline(file, s, '\n');	
			if (file.fail()) {
				break; // ok, got to the end of this file (we hope)
				
			} else if (s == "<MARKER>") {
				file >> markerNameStr >> chrNumStr >> locInBases >> locIn_cM;
				//cout << markerNameStr << ", " << chrNumStr << ", " << locInBases << ", " << locIn_cM << endl; // for debugging... see what markers are read in
				assert(chrNumStr.size() >= 1);
				if (chrNumStr[0] == 'x' || chrNumStr[0] == 'X') {
					throw new string("We found a marker on the X chromosome.\nWhen labeling markers, please just give the chromosome a number,\nso that X is the next-to-last chomosome,\nand Y is the last chromosome.\nFor mice, we would have X=20 and Y=21.");
					//interpretedChrIndex = Chromosome::xChrIndex();
					
				} else if (chrNumStr[0] == 'y' || chrNumStr[0] == 'Y') {
					throw new string("We found a marker on the Y chromosome.\nWhen labeling markers, please just give the chromosome a number,\nso that X is the next-to-last chomosome,\nand Y is the last chromosome.\nFor mice, we would have X=20 and Y=21.");
					//interpretedChrIndex = Chromosome::yChrIndex();
					
				} else {
					interpretedChrIndex = atoi(chrNumStr.c_str()) - 1; // all but X and Y chrs
					
					if (interpretedChrIndex < 0) {		    
						throw new string(string("Error in MouseIO::getMarkerDataFromFile. One of the markers in the marker file\nisn't on a valid chromosome (it claims to be on chromosome \"") + chrNumStr + string("\"."));
					}
				}
				
				newCollection->addMarker(interpretedChrIndex, markerNameStr, locInBases, locIn_cM);
				// selection coefficients aren't done yet
			}
		}
		//cout << "NOTE: Reading markers: selection coefficients aren't done yet\n";
		file.close();
		
		return newCollection;
		
	} catch (const string *e) {
		throw e;
	}
}




// ====================================================================================================================
// ====================================================================================================================
char MouseIO::nucleotideToABH(GenomeMixer *mixPtr, int theNum) {
	return MouseIO::nucleotideToABH(mixPtr, theNum, theNum);
}

char MouseIO::nucleotideToABH(GenomeMixer *mixPtr, int matNum, int patNum) {
	// This returns either an 'A', a 'B', or an 'H' for hybrid, depending on the parental values.
	// If you give it a "matNum" or "patNum" that corresponds to a purebred strain whose symbol is not 'A' or 'B',
	// then it throws an error.
	
	try {
		char matChar = mixPtr->getPurebredSymbol(matNum)[0];
		char patChar = mixPtr->getPurebredSymbol(patNum)[0];
		// Above: Note that we are just getting the first character of a string.
		// This won't be right if the symbols are multiple digits, like "103" and "105," since our program
		// will think they are the same for this kind of export. Fortunately, anyone who calls "nucleotideToABH"
		// expects only an A, B, or H, and won't handle anything else, so the output wouldn't work anyway.
		
		if (matChar != patChar) {
			return 'H'; // hybrid
		} else {
			// We want to print an 'A' or a 'B'. If we get some other letter, give a warning.
			if (matChar != 'A' && matChar != 'B') {			
				cerr << "In this kind of output, we can only deal with two parental strains (AA and BB). But it looks like we have more parental strains in this output, so the file format won't be correct. The parental symbls were \"" <<  mixPtr->getPurebredSymbol(matNum) << "\" for the female, and \"" << mixPtr->getPurebredSymbol(patNum) << "\" for the male." << endl;
				throw new string("In this kind of output, we can only deal with two parental strains (AA and BB). But it looks like we have more parental strains in this output, so the file format won't be correct.");
			}
			return matChar;
		}
		
	} catch (const string *e) {
		throw e;
	}
}


// ====================================================================================================================
// ====================================================================================================================
// ******** ******** ******** ******** ******** ******** ******** ******** ********
//	    Write Files in QGene format (new) (map and population)
// ******** ******** ******** ******** ******** ******** ******** ******** ********

void MouseIO::writeQGeneFiles(string fileToWriteStr, GenomeMixer *mixPtr, string endLineStr) {
	// endLineStr is the format of the linebreaks. For Mac OS 9, it was "\r", for UNIX it's "\n" (also for OS X).
	// The endLineStr format is specified here so that we can output Mac OS 9 linebreaks for QGene, even if we are running
	// on a UNIX machine (QGene only runs on OS 9, for now).
	
	try {
		assert(mixPtr != NULL && mixPtr->markerColPtr != NULL);
		MarkerCollection *markerDataPtr = mixPtr->markerColPtr;
		
		// We don't write out "fileToWriteStr" exactly as it is passed in;
		// instead, we write out "fileToWriteStr" + ".pop" and "fileToWriteStr" + ".map" for the population and map files.
		
		string dataType     = "f2";   // 'HxHf1';
		string charCodes    = "ABHXY-"; // <-- the valid symbols... A, B, Hybrid, X, Y, and "nothing"
		long numTraits      = 0; // not used yet
		int numIndividuals  = mixPtr->numMiceToExport(); // the number of mice that will be printed
		
		// "Standard population types are:       <--  From Clare Nelson's QGene web site (www.qgene.org)
		//	 dhf1	doubled haploids from F1         //	ri	recombinant inbred
		//	bc1f1	backcross of F1                  //	bc2f1	second backcross following F1
		//	bc3f1	third backcross following F1     //	bc1f2	backcross of F2
		//	bc2f2	second backcross following F2    //	f2	F2
		//	f3	F3                               //	bc1s1	selfed bc1f1
		//	bc1s2	twice-selfed bc1f1               //	bc2s1	selfed bc2f1
		//	hxhf1	F1 of two possibly heterozygous parents
		// Marker genotype codes used in the datafile represent homozygote AA, homozygote aa,
		// heterozygote Aa, ambiguous dominant markers a_ and A_, and missing data.
		// For those familiar with Mapmaker, this is the same scheme.
		// An additional 9 character codes (not shown here) have been added in version 3.0
		// for crosses with up to four alleles segregating, designated by type HxHf1.
		
		if (numIndividuals <= 0) {
			cerr << "No mice were set to be output, so we can't write QGene output." << endl
				<< "You must specify cells to export, using the \"exp\" keyword.";
			throw new string("Can't write QGene output: No mice were set to be output\n(Use the \"exp\" keyword to specify cells to export).");
		} else if (numIndividuals > 2048) {
			cerr << "Can't write QGene output: A dataset (for export to QGene)" << endl
				<< "may only have as many as 2048 individuals exported to it." << endl
				<< "Unfortunately, we have " << numIndividuals << " individuals.";
			throw new string("Can't write QGene output: Too many individuals (max of 2048)");
		}
		
		if (markerDataPtr->totalMarkers() > 2048) {
			cerr << "Can't write QGene output: A dataset (for export to QGene)" << endl
				<< "may only have as many as 2048 markers exported to it." << endl
				<< "However, we have a total of " << markerDataPtr->totalMarkers() << " markers.";
			throw new string("Can\'t write QGene output: too many markers (max of 2048)");
		}
		
		if (numTraits > 500) {
			cerr << "Can't write QGene output: A dataset (for export to QGene)" << endl
				<< "may only have as many as 500 traits exported to it." << endl
				<< "However, we have a total of " << numTraits << " traits.";
			throw new string("Can't write QGene output: too many traits (max of 500)");
		}
		
		ofstream mapfile;
		mapfile.open(string(fileToWriteStr + string(".map")).c_str(), ios::out); //  | ios::noreplace
		if (mapfile.fail()) {// || !mapfile.is_open()) {
			cerr << "Can't write QGene output: We were trying to write" << endl
				<< "to this location: " << fileToWriteStr << ".map" << endl
				<< ", but some kind of error occurred, and we couldn't open the file.";
			throw new string("Can't write QGene output: Some sort of error occured, and we couldn't open the output file.");
		}
		
		// Now we are writing to the file...
		mapfile << "Data type " << dataType << endLineStr.c_str();
		mapfile << numIndividuals << ' ' << markerDataPtr->totalMarkers() << ' ' << numTraits << ' ' << markerDataPtr->markersForHowManyChr() << ' ' << charCodes << endLineStr.c_str();
		mapfile << "Parents A B" << endLineStr.c_str();
		
		//	data type XYZ
		//	number_of_individuals		number_of_markers		num_traits		num_chromosomes_in_animal		valid_data_at_each_marker
		//	parents a b
		//	*chromosomeName		num_markers-this-chr-only	markername	dist	markername	...
		
		
		// Just get the first mouse... it will be representative, as far as number of chromosomes
		Mouse *oneMouse = mixPtr->mouseToExport(0);
		
		// loop for each chromosome
		for (int chr = 0; chr < markerDataPtr->numberOfChrs(); chr++) {
			long markersInThisChr = markerDataPtr->markersInChr(chr);
			if (!markersInThisChr) {
				// There are no markers to write, so we don't output this chromosome to QGene.
				// QGene isn't interested in chromosomes with no markers.
				continue;
			}
			// Otherwise, there are actually markers to write...
			mapfile << '*' << oneMouse->getChr(kMALE, chr)->name() << '\t' << markersInThisChr << ' ';
			
			// i is the marker index for this chromosome only. Goes from 0 to markersThisChr-1.
			for (int i = 0; i < markersInThisChr; i++) {
				
				mapfile << markerDataPtr->name(chr, i);
				
				if (i+1 < markersInThisChr) {
					// Prints the distance (in cM! Not in nucleotides!) between this marker and the next one (but only if there *is* a next marker)
					float distanceBetween_cM    =    markerDataPtr->cM(chr, i+1) - markerDataPtr->cM(chr, i);
					mapfile << " " << distanceBetween_cM << " ";		
					if (distanceBetween_cM <= 0) {
						// Non-fatal error
						cerr << ">>> Warning: Error in QGene map file output. A pair of markers on the chromosome \""
							<< Chromosome::chrNameFromIndex(chr) << "\" is out of order." << endl;
						cerr << "     (Markers with indices " << i << " and " << (i+1) << ")." << endl;
						cerr << "     Marker locations should be given in increasing order in the marker input file." << endl;
					}
				}
			}
			mapfile << endLineStr.c_str();
		}
		
		// chr   nummarkers  marker distance marker distance etc...
		//print WRITEMAP "*Chr1	3 chr1marker1_name	5.0	chr1marker2_name	5.0	chr1marker3_name\n";
		//print WRITEMAP "*Chr2	3 chr2marker1_name	5.0	chr2marker2_name	5.0	chr3marker3_name\n";	
		mapfile.close();	
		
		ofstream popfile;
		popfile.open(string(fileToWriteStr + string(".pop")).c_str(), ios::out); //  | ios::noreplace
		if (popfile.fail() || !popfile.is_open()) {
			cerr << "Error: we were trying to write breeding scheme file output to the file \"" << fileToWriteStr << ".pop\", but an error of some kind occurred," << endl << "and we couldn't open the file." << endl;
			throw new string("Error while writing breeding scheme file. Some sort of error occurred, and we couldn't open it.");
		}
		
		// writing population file		writing population file		writing population file
		long markIndexThisChr = 0;
		int  chr = 0;
		long markerCount = 0;
		
		
		while (markerCount < markerDataPtr->totalMarkers()
			&& chr < markerDataPtr->numberOfChrs() 
			&& chr < Chromosome::numChr() ) {
			
			// go through all the markers, on all the chromosomes (that have markers, that is)
			long markersInThisChr = markerDataPtr->markersInChr(chr);
			long locOfThisMarker = markerDataPtr->base(chr, markIndexThisChr);
			
			if (!markersInThisChr) {
				// in case this chromosome doesn't have any markers...
				chr++;  // next, please...
				markIndexThisChr = 0;
			} else {
				// Otherwise, this is a valid marker...
				popfile << '*' << markerDataPtr->name(chr, markIndexThisChr) << '\t';
				// now, for each individual, get the data at that marker point, specified by chromosome and "gene" location
				bool validIndex = true;
				
				for (int i = 0; i < mixPtr->numMiceToExport(); i++) {
					
					Mouse *currentCreature = mixPtr->mouseToExport(i);
					
					if (validIndex &&
						locOfThisMarker <= currentCreature->getChr(kMALE, chr)->lengthBases() &&
						locOfThisMarker >= 0) {
						// This marker is actually in the range of the chromosome...
						if (chr == Chromosome::xChrIndex() && currentCreature->isMale()) {
							// male (only one X chromosome)
							int matValue =  currentCreature->geneAtBase(kFEMALE, chr, locOfThisMarker);
							popfile << MouseIO::nucleotideToABH(mixPtr, matValue);
							
						} else if (chr == Chromosome::yChrIndex()) {
							if (currentCreature->isFemale()) {
								popfile << '-'; // female (no Y chromosome)
							} else {
								// male (so include the Y chromosome)
								int patValue = currentCreature->geneAtBase(kMALE, chr, locOfThisMarker);
								popfile << MouseIO::nucleotideToABH(mixPtr, patValue);
							}
							
						} else if (chr < 0 || chr >= Chromosome::numChr()) {
							cerr << "Error while outputting QGene-format files: " << endl 
								<< "Trying to write to a chromosome that appears to not exist (the offending index is " << chr << ").";
							throw new string("Errow while outputting QGene-format files. Tried to write a non-existing chromosome.");
							
						} else {		
							int matValue =  currentCreature->geneAtBase(kFEMALE, chr, locOfThisMarker);
							int patValue =  currentCreature->geneAtBase(kMALE,   chr, locOfThisMarker);
							popfile << MouseIO::nucleotideToABH(mixPtr, matValue, patValue);
						}
						
					} else if (validIndex) {
						// Now this marker is outside of the range of the chromosome!
						// Non-fatal error
						validIndex = false;
						cerr << ">>> Warning: Did not write a marker on " << Chromosome::chrNameFromIndex(chr) << ", marker with index $markIndexThisChr (marker number " << markerCount << " overall). This item was outside of the range of the chromosome." << endl;		    
					} else {
						// nothing; don't keep reporting "out of bounds" errors for the same chr over and over
					}
					
				} // end of "for (int i = 0; i < mixPtr->numMiceToExport(); i++) {"
				
				popfile << endLineStr.c_str(); // <-- we're done printing that marker data, now for the next one
				
				markIndexThisChr++; // maybe there are more markers for this chromosome
				markerCount++;
				
				if (markIndexThisChr >= markersInThisChr) {
					markIndexThisChr = 0;  // there are no more markers on this chromosome,
					chr++;  // so now we reset and move to the next one
				}
				// end of "if valid"
			}
			
		} // end of "while" loop
		
		// Example of marker data:
		// marker  data for each individual (Mouse1marker1 Mouse2marker1, etc...)
		//      print WRITEPOPULATION "*Fish	11221\n";
		//      print WRITEPOPULATION "*Squid	12111\n";
		//      print WRITEPOPULATION "*Octopus	12121\n";
		//      print WRITEPOPULATION "*Fish2	21221\n";
		//      print WRITEPOPULATION "*Squid2	12221\n";
		//      print WRITEPOPULATION "*Octopus2	21221\n";
		
		popfile.close();
		
		
	} catch (const string *e) {
		throw e;
	}
}


// =========================================================================================================
// =========================================================================================================
// ******** ******** ******** ******** ******** ******** ******** ******** ********
//			Write Files in MapManager text format
// ******** ******** ******** ******** ******** ******** ******** ******** ********

void MouseIO::writeMapManagerFormat(string fileToWriteStr, GenomeMixer *mixPtr, string endLineStr) {
	// endLineStr is the format of the linebreaks. For Mac OS 9, it was "\r", for UNIX it's "\n" (also for OS X), and for windows it's both of them together in some order that I forget
	try {
		assert(mixPtr != NULL && mixPtr->markerColPtr != NULL);
		MarkerCollection *markerDataPtr = mixPtr->markerColPtr;
		
		int numMarkersTotal = markerDataPtr->totalMarkers();
		int numIndividuals	   = mixPtr->numMiceToExport();	// <-- number to be printed
		
		ostringstream stringStream;
		stringStream << fileToWriteStr << "_(n=" << numIndividuals << ')';
		fileToWriteStr = stringStream.str();
		
		if (numIndividuals<= 0) {
			throw new string("Error in writing MapManager Files: No individuals were set to be output.\nYou must specify cells to export, using the \"exp\" keyword.");
		}
		
		ofstream mmfile;
		mmfile.open(fileToWriteStr.c_str(), ios::out);
		if (mmfile.fail()) {// || !mmfile.is_open()) {
			cerr << "Error in writing MapManager files: failed to open the file." << endl
				<< "File was: \"" << fileToWriteStr << "\"";
			throw new string("Error in writing MapManager files: failed to open the output file for writing.");
			
		}
		
		int markIndexThisChr = 0;
		int chr = 0;
		int markerCount = 0;
		
		// Write the top line, one line for each column of data
		// So... this means, we write the number of individuals (each marker data item is considered a column,
		// so "AABBA" is 5 columns), plus the number of columns for header data, like which chromosome the
		// marker is on, and what the marker name is.
		
		for (int i = 1; i <= numIndividuals+4; i++) {
			mmfile << i << ' ';
		}
		mmfile << endLineStr.c_str();
		
		while ((markerCount < numMarkersTotal)
			&& 	(chr < markerDataPtr->numberOfChrs())
			&& 	(chr < Chromosome::numChr()) ) {
			// go through all the markers, on all the chromosomes (the ones that have markers, that is)
			
			long markersInThisChr = markerDataPtr->markersInChr(chr);
			long locOfThisMarker   = markerDataPtr->base(chr, markIndexThisChr);
			
			if (markersInThisChr <= 0) { 
				chr++; // if this chromosome doesn't have any markers, go to the next one
				markIndexThisChr = 0;
				
			} else {
				// Otherwise, this is a valid marker...
				// Writing format: Import	1	D1Mit211	HDDHDDHDDDDUD
				
				//  Print the rows of data here..
				mmfile 	    << "GenomeMixer" << '\t';
				if (chr == Chromosome::xChrIndex())       mmfile << 'X';
				else if (chr == Chromosome::yChrIndex()) mmfile << 'Y';
				else                                                            mmfile << (chr+1); // plain autosome
				
				mmfile 	<< '\t' //  writes which chromosome the marker is on (the three lines of code readabove), then a tab (this line)
					<< markerDataPtr->name(chr, markIndexThisChr) << '\t' // <-- name of this marker
					<< markerDataPtr->cM(chr, markIndexThisChr) << '\t'; // <-- MapManager "alias" field (distance in cM)
				
				// Now, for each individual, get the data at that marker point, specified by chromosome and "gene" location
				// (this is the "BBBAAABBHHABHABABAHHBA" part of the row)
				bool validIndex = true;
				for (int i = 0; i < numIndividuals; i++) {
					
					Mouse *currentCreature = mixPtr->mouseToExport(i);
					//cout << currentCreature->name() << " is being printed to MapManager QTX format." << endl;		
					if (validIndex
						&& locOfThisMarker <= currentCreature->getChr(kMALE, chr)->lengthBases()
						&& locOfThisMarker >= 0) {
						// This marker is actually in the range of the chromosome...
						if (chr == Chromosome::xChrIndex() && currentCreature->isMale()) {
							// male (only one X chromosome)
							int matValue =  currentCreature->geneAtBase(kFEMALE, chr, locOfThisMarker);
							mmfile << MouseIO::nucleotideToABH(mixPtr, matValue);
							
						} else if (chr == Chromosome::yChrIndex()) {
							if (currentCreature->isFemale()) {
								// female (no Y chromosome)
								mmfile << '-';
							} else {
								// male
								int patValue =  currentCreature->geneAtBase(kMALE, chr, locOfThisMarker);
								mmfile << MouseIO::nucleotideToABH(mixPtr, patValue);
							}
							
						} else if (chr < 0 || chr >= Chromosome::numChr()) {
							throw new string("Error: Trying to write a chromosome that appears to not exist.");
							
						} else {
							int matValue =  currentCreature->geneAtBase(kFEMALE, chr, locOfThisMarker);
							int patValue =  currentCreature->geneAtBase(kMALE, chr, locOfThisMarker);
							mmfile << MouseIO::nucleotideToABH(mixPtr, matValue, patValue); 
						}
						
					} else {
						// This marker is outside of the range of the chromosome!
						if (validIndex) {
							// Non-fatal error
							cerr << ">>> Warning: Did not write a marker on "  << Chromosome::chrNameFromIndex(chr)  << ". This item was outside the range of the chromosome." << endl;
							validIndex = false;
						}
					}
				} // end of "for" loop
				
				mmfile << endLineStr.c_str(); // <-- we're done printing that marker data, now for the next one
				
				markIndexThisChr++;	// maybe there are more markers for this chromosome
				markerCount++;
				
				if (markIndexThisChr >= markersInThisChr) {
					markIndexThisChr = 0;	// there are no more markers on this chromosome,
					chr++;			// so now we reset and move to the next one
				}
				// end of "if valid"
			} // end of "else"
			
		} // end of "for" loop    
		mmfile.close();
		
	} catch (const string *e) {
		throw e;
	}
}








// =========================================================================================================
// =========================================================================================================
// ******** ******** ******** ******** ******** ******** ******** ******** ********
//			Write Files in R/qtl text format
// ******** ******** ******** ******** ******** ******** ******** ******** ********

void MouseIO::writeRQtlFormat(string fileToWriteStr, GenomeMixer *mixPtr, string endLineStr) {
	try {
		cout << "Writing QTL format: Note that we currently don't export the Y chromosomes." << endl;
		
		// endLineStr is the format of the linebreaks. For Mac OS 9, it was "\r", for UNIX it's "\n" (also for OS X), and for windows it's both of them together in some order that I forget
		
		// R/qtl likes comma-delimited files in the following format:
		// Phenotype1  ...   MarkerChr1_a   MarkerChr1_b ... MarkerChr84_f ... MarkerChrX_a MarkerChrX_b
		// (nothing)   ...   0 (in cM)         32.41             39               0          24
		// 1232 kg     (each line here is an individual mouse/animal)
		// 39 kg       ...      A              H                 B                H          A
		// 41 kg                B              H                 B                A          A
		//  ... and so on, with the traits on the left, and then the marker data to the right
		
		// From the R/qtl documentation:
		// "The input file is a text file with a specified field delimiter (sep) (a comma is recommended). 
		// The first line should contain the phenotype names followed by the marker names.  At least one phenotype must be included; for example, include a numerical index for each individual. 
		// The second line should contain blanks in the phenotype columns, followed by chromosome identifiers for each marker in all other columns. If a chromosome has the identifier X or x, it is assumed to be the X chromosome; otherwise, it is assumed to be an autosome. 
		// An optional third line should contain blanks in the phenotype columns, followed by marker positions, in cM. 
		// Marker order is taken from the cM positions, if provided; otherwise, it is taken from the column order. 
		// Subsequent lines should give the data, with one line for each individual, and with phenotypes followed by genotypes.  If possible, phenotypes are made numeric; otherwise they are converted to factors."
		
		if (mixPtr == NULL || mixPtr->markerColPtr == NULL) {
			throw new string("Error in writing R/qtl format files:\nNo GenomeMixer has been defined yet.\nSpecify settings before trying to export this format.");
		}
		MarkerCollection *markerDataPtr = mixPtr->markerColPtr;
		
		const int phenIDIndex  = 0; // <------------+
		const int phenSexIndex = 1; // <------------|----------------|
		const int numberOfFixedPhenotypes = 2; // one is ID, the other is sex
		
		int numChrsToPrint = Chromosome::numChr() - 1; // <-- don't print the Y chromosome
		int numIndividuals  = mixPtr->numMiceToExport();	// <-- number of mice (rows, here) to be printed
		int numPhenotypes   = /*mixPtr->numPhenotypes() + */ + numberOfFixedPhenotypes; // right now, we have no phenotypes!
		if (numIndividuals<= 0) {
			throw new string("Error in writing R/qtl format files:\nNo individuals were set to be output.\nYou must specify cells to export, using the \"exp\" keyword.");
		}
		
		ofstream rqtlout;
		rqtlout.open(fileToWriteStr.c_str(), ios::out); // ios::nocreate
		if (rqtlout.fail()) { // || !rqtlout.is_open()) {
			cerr << "Error in writing R/qtl files: failed to open the file." << endl
				<< "File was: \"" << fileToWriteStr << "\"";
			throw new string("Error in writing R/qtl files:\nfailed to open the output file for writing.");
		}
		
		// --------- WRITE OUT THE FIRST LINE: PhenotypeName,PhenotypeName,MarkerName,MarkerName...
		for (int phen = 0; phen < numPhenotypes; phen++) {
			switch (phen) {
			case phenIDIndex:
				rqtlout << "Mouse_ID" << ',';
				break;
			case phenSexIndex:
				rqtlout << "Sex" << ',';
				break;
			default:
				rqtlout << "PhenotypeName" << ',';
				break;
			}
		}
		
		for (int chr = 0; chr < numChrsToPrint; chr++) {
			assert (Chromosome::yChrIndex() != chr);
			
			long markersThisChr = markerDataPtr->markersInChr(chr);
			for (long mark = 0; mark < markersThisChr; mark++) {
				rqtlout << markerDataPtr->name(chr, mark).c_str();
				if (chr+1 < numChrsToPrint || mark+1 < markersThisChr) { rqtlout << ','; } // we can't have a comma at the end of a line, or R/qtl is sad
			}
		}
		rqtlout << endLineStr.c_str();
		
		
		// --------- WRITE OUT THE SECOND LINE: ,,,,,MarkerOnChr1,MarkerOnChr1,...,MarkerOnChrX
		// Now, write the second line: for each phenotype, we just give nothing. For each marker, write its chromosome number, or X. I don't think we support Y here, so just skip it.
		for (int phen = 0; phen < numPhenotypes; phen++) {      
			rqtlout << ',';
		}    
		for (int chr = 0; chr < numChrsToPrint; chr++) {
			
			if (Chromosome::yChrIndex() == chr)
				continue; // don't output the y chromosome in this format
			
			long markersThisChr = markerDataPtr->markersInChr(chr);
			for (long mark = 0; mark < markersThisChr; mark++) {
				if (chr == Chromosome::xChrIndex()) { rqtlout << 'X'; }
				else                                { rqtlout << (chr+1); }	
				
				if (chr+1 < numChrsToPrint || mark+1 < markersThisChr) { rqtlout << ','; } // we can't have a comma at the end of a line, or R/qtl is sad
			}
		}
		rqtlout << endLineStr.c_str();
		
		// --------- WRITE OUT THE THIRD LINE: ,,,,,MarkerPositionIn_cM,MarkerPositionIn_cM,...
		for (int phen = 0; phen < numPhenotypes; phen++) {      
			// just output commas for the phenotype
			rqtlout << ','; // commas, but not at end of line
		}    
		for (int chr = 0; chr < numChrsToPrint; chr++) {
			long markersThisChr = markerDataPtr->markersInChr(chr);
			for (long mark = 0; mark < markersThisChr; mark++) {
				rqtlout << markerDataPtr->cM(chr, mark);	
				if (chr+1 < numChrsToPrint || mark+1 < markersThisChr) { rqtlout << ','; } // we can't have a comma at the end of a line, or R/qtl is sad
			}
		}
		rqtlout << endLineStr.c_str();
		
		
		// WRITE LINES FOUR AND BEYOND (the invidual mouse data)
		// The rest of what we do is just write: "Individual_Phenotype1_value,Individual_Phenotype2_value,...,Marker_value(A,B,H...),Marker_value(A,B,H...), etc...
		
		for (int indiv = 0; indiv < numIndividuals; indiv++) {      
			Mouse *currentCreature = mixPtr->mouseToExport(indiv);
			
			for (int phen = 0; phen < numPhenotypes; phen++) { // Output this mouse's phenotype data	
				switch (phen) {
				case phenIDIndex:
					rqtlout << currentCreature->ID();
					break;
				case phenSexIndex:
					if (currentCreature->isFemale()) { rqtlout << 'f'; }
					else                             { rqtlout << 'm'; }
					break;
				default:
					rqtlout << "123.456"; // <-- we don't have phenotype data for now
					break;
				}
				rqtlout << ',';
			}
			
			for (int chr = 0; chr < currentCreature->numChr() && chr < numChrsToPrint; chr++) { // Output this mouse's genotype data (as 'A' 'B' or 'H')
				long markersThisChr = markerDataPtr->markersInChr(chr);
				for (long mark = 0; mark < markersThisChr; mark++) {	  
					long locOfThisMarker = markerDataPtr->base(chr, mark);
					int  matValue = currentCreature->geneAtBase(kFEMALE, chr, locOfThisMarker);
					
					if (Chromosome::xChrIndex() == chr && currentCreature->isMale()) {
						// Male: only one X chromosome
						rqtlout << MouseIO::nucleotideToABH(mixPtr, matValue);
						
						//rqtlout << mixPtr->getPurebredSymbol(matValue) << ':' << '-'; // no paternal data, since this is the X chr on a male
						
					} else {
						int patValue = currentCreature->geneAtBase(kMALE  , chr, locOfThisMarker);	  
						rqtlout << MouseIO::nucleotideToABH(mixPtr, matValue, patValue);
						
						//rqtlout << mixPtr->getPurebredSymbol(matValue) << ':' << mixPtr->getPurebredSymbol(patValue);
					}
					
					if (chr+1 < numChrsToPrint || mark+1 < markersThisChr) { rqtlout << ','; } // we can't have a comma at the end of a line, or R/qtl is sad
				}
			}
			rqtlout << endLineStr.c_str();
		} // end of "for all individuals" loop
		
		rqtlout.close();
		
		if (rqtlout.fail()) { // || rqtlout.is_open()) {
			throw new string("Error in writing R/qtl files: failed to <B>close</B> the output file for writing.");
		}
		
		
	} catch (const string *e) {
		throw e;
	}
} // END OF WRITING R/qtl format





// ----------------------------- WRITE MARKER FILE ---------------------------------------------------------
void MouseIO::writeMarkerFile(string fileToWriteStr, GenomeMixer *mixPtr, string endLineStr) {    
	
	try {
		if (mixPtr == NULL || mixPtr->markerColPtr == NULL) {
			cerr << "MouseIO::writeMarkerFile was called with a null pointer.";
			assert(mixPtr != NULL && mixPtr->markerColPtr != NULL);
		}
		
		ofstream outfile;
		outfile.open(fileToWriteStr.c_str(), ios::out); // ios::nocreate
		if (outfile.fail() || !outfile.is_open()) {
			cerr << "Error in writing our markers: failed to open the file." << endl
				<< "File was: \"" << fileToWriteStr << "\"";
			throw new string("Error in writing our marker file:\nfailed to open the output file for writing.");
		}
		
		outfile << "This is a marker location file." << endLineStr.c_str()
			<< "Comments can go here at the top, but they will be lost if the file is saved again." << endLineStr.c_str()
			<< "Note that spaces are *not* allowed in marker names. Use underscores or dashes instead, if you need to." << endLineStr.c_str()
			<< endLineStr.c_str() << endLineStr.c_str() << endLineStr.c_str()
			<< "When this file was first saved, it had a total of "
			<< mixPtr->markerColPtr->totalMarkers() << " marker (counted by GenomeMixer)." << endLineStr.c_str()
			<< "-----------------------real marker data below---------------------------------------" << endLineStr.c_str();
		
		outfile << "<MARKER NAME>\t<CHROMOSOME NUMBER>\t<LOCATION in Bases>\t<LOCATION in cM (centimorgans>\t<PROB OF SURVIVAL>" << endLineStr.c_str();
		outfile << "<START MARKERS>" << endLineStr.c_str();
		
		int markerCount = 0;
		
		for (int chrIndex = 0;   chrIndex < mixPtr->markerColPtr->numberOfChrs();   chrIndex++) {
			for (long markerOnChr = 0;   markerOnChr < mixPtr->markerColPtr->markersInChr(chrIndex);   markerOnChr++) {
				outfile << "<MARKER>"
					<< '\t';
				outfile << (mixPtr->markerColPtr->name(chrIndex, markerOnChr)).c_str()
					<< '\t';
				outfile << (chrIndex + 1)
					<< '\t';	    
				outfile << mixPtr->markerColPtr->base(chrIndex, markerOnChr)
					<< '\t';
				outfile << mixPtr->markerColPtr->cM(chrIndex, markerOnChr)
					<< endLineStr.c_str();
				markerCount++;
			}
		}
		
		if (markerCount != mixPtr->markerColPtr->totalMarkers()) {
			cerr << "Strangely, we didn't output the correct number of markers (markerCount was: " 
				<< markerCount  << ", totalMarkers() was: "		
				<< mixPtr->markerColPtr->totalMarkers() << ")" << endl;
		}    
		
		outfile.close();
		
	} catch (const string *e) {
		throw e;
	}
}



// ----------------------------- SETTINGS FILE ---------------------------------------------------------
void MouseIO::writeSettingsFile(string fileToWriteStr, GenomeMixer *mixPtr, string endLineStr) {
	
	
	try { 
		
		if (mixPtr == NULL) {
			throw new string("Programming Error: MouseIO::writeSettingsFile was\ncalled with a null pointer.");
		}
		
		ofstream outfile;
		outfile.open(fileToWriteStr.c_str(), ios::out); // ios::nocreate
		if (outfile.fail() || !outfile.is_open()) {
			throw new string(string("Error in writing our settings:\nfailed to open the file \"") + fileToWriteStr + string("\" for output."));
		}
		
		outfile << "<NAME>" << '\t' << fileToWriteStr.c_str() << endLineStr.c_str()
			<< "<VERSION>" << '\t' << mixPtr->version << endLineStr.c_str()
			<< endLineStr.c_str();
		
		for (int i = 0; i < Chromosome::numChr(); i++) {
			if (Chromosome::xChrIndex() == i) {
				outfile <<"<FEMALE_SEX_CHR>";
			} else if (Chromosome::yChrIndex() == i) {
				outfile << "<MALE_SEX_CHR>";
			} else {
				outfile << "<CHR>" ;
			}
			outfile << '\t' 
				<< (Chromosome::chrNameFromIndex(i)).c_str() << ' '
				<< Chromosome::lengthMBASEfromIndex(i) << " MB "
				<< Chromosome::length_cM_fromIndex(i) << " cM"
				<< endLineStr.c_str();
		}
		
		outfile << endLineStr.c_str();
		outfile << "<DEFAULT_LITTER_SIZE>" << '\t' << mixPtr->defaultLitterSize << endLineStr.c_str();
		outfile << "<DEFAULT_LITTER_STDEV>" << '\t' << mixPtr->defaultLitterStDev << endLineStr.c_str();

		outfile << "<ABSOLUTE_INTERFERENCE>" << '\t' << mixPtr->noRecombDistBases << endLineStr.c_str();
		
		outfile << "<RECOMBINATION_MODEL>" << '\t' << mixPtr->recombModel << '\t' << "(0 = none, 1 = linear, 2 = Haldane)" << endLineStr.c_str();

		outfile << endLineStr.c_str();
		for (int i = 0; i < mixPtr->numParentalStrains(); i++) {
			outfile << "<PUREBRED>" << '\t' << (mixPtr->getPurebredName(i)).c_str() << '\t' << (mixPtr->getPurebredSymbol(i)).c_str() << endLineStr.c_str();
		}
		
		outfile << "<GENERATIONS " << mixPtr->numGenerations() << '>';
		for (int i = 0; i < mixPtr->numGenerations(); i++) {
			outfile << '\t' << (i+1);
		}
		outfile << endLineStr.c_str();
		
		for (int cage = 0; cage < mixPtr->numCages(); cage++) {
			outfile << "<CAGE " << (cage+1) << '>';
			for (int gen = 0; gen < mixPtr->numGenerations(); gen++) {
				
				outfile << '\t' << '\"';
				if (mixPtr->getCellData(cage, gen)) {
					// If there is anything in the cell, print it.
					// Otherwise, we just have nothing between quotes: ""
					outfile << (mixPtr->getCellData(cage, gen)->toString(mixPtr->purePtr())).c_str();
				}
				outfile << '\"';
			}
			outfile << endLineStr.c_str();
		}
		
		outfile << "<END_DATA>" << endLineStr.c_str();
		outfile << "==============================================================================================" << endLineStr.c_str();
		outfile << "-----------------------end of real data: comments below---------------------------------------" << endLineStr.c_str();
		
		outfile << "Comments can go here, but not up with the data (and not above the data, either), or the file won't be readable by GenomeMixer." << endLineStr.c_str();
		
		outfile.close();
		
	} catch (const string *e) {
		throw e;
	}
}


